/*
 * Actions for parser for logic language
 */

#include <stdio.h>
#include <stdarg.h>
#include "logsyn.tab.h"
#include "logsyn.h"

#define swap(a, b, t)	{ t = a; a = b; b = t; }

int dec_level;			/* current level of symbol declarations */
int num_params;			/* # params so far in function defn */

struct expr *inputs;		/* group of inputs seen so far */
struct expr *outputs;		/* group of outputs seen so far */
int num_signals;		/* total # signals seen so far */

int n_errs;			/* # errors encountered */

int ddebug = 0;			/* debug printout on vble definitions */

extern FILE *yyin;
int lineno = 1;

/* Stack of suspended include files */
#define INCLMAX	10

struct incl_file {
	char	*name;
	FILE	*susp;
	int	lineno;
} incl_stack[INCLMAX];
int incl_level;

extern char *input_file;

where()
{
	fprintf(stderr, "%s: %d: ",
		incl_level > 0? incl_stack[incl_level].name: input_file,
		lineno);
}

yyerror(char *s, ...)
{
	va_list args;

	where();
	va_start(args, s);
	vfprintf(stderr, s, args);
	va_end(args);
	fprintf(stderr, "\n");
	++n_errs;
}

yywrap()
{
	if( incl_level > 0 ){
		--incl_level;
		fclose(yyin);
		yyin = incl_stack[incl_level].susp;
		lineno = incl_stack[incl_level].lineno;
		return 0;
	}
	return 1;
}

do_include(str)
char *str;
{
	FILE *f;
	int l;

	if( incl_level >= INCLMAX-1 ){
		yyerror("too many levels of include files");
		free(str);
		return;
	}
	l = strlen(str);
	if( l >= 2 ){
		str[l-1] = 0;		/* knock off " */
		f = fopen(str+1, "r");
		if( f != NULL ){
			incl_stack[incl_level].susp = yyin;
			incl_stack[incl_level].lineno = lineno;
			++incl_level;
			incl_stack[incl_level].name = strsave(str + 1);
			yyin = f;
			lineno = 1;
			free(str);
			return;
		}
		str[l-1] = '"';
	}
	yyerror("include file %s not found", str);
	free(str);
}

/*
 * Storage management
 */
char *
alloc(n)
int n;
{
	char *p;

	p = malloc(n);
	if( p == NULL ){
		fprintf(stderr, "Out of memory!\n");
		exit(1);
	}
	return p;
}

char *
strsave(s)
char *s;
{
	char *p;

	p = alloc(strlen(s) + 1);
	strcpy(p, s);
	return p;
}

/*
 * Symbol table
 */

#define SYMSIZE	64
struct ident *symtab[SYMSIZE];

hash(name)
char *name;
{
	int h;

	h = 0;
	while( *name != 0 )
		h += *name++;
	h %= SYMSIZE - 1;
	return h;
}

struct ident *
lookup(name)
char *name;
{
	struct ident *ip;

	ip = symtab[hash(name)];
	while( ip != NULL ){
		if( strcmp(ip->i_name, name) == 0 )
			break;
		ip = ip->i_next;
	}
	return ip;
}

enter(ip)
struct ident *ip;
{
	struct ident **hp;

	hp = &symtab[hash(ip->i_name)];
	ip->i_next = *hp;
	ip->i_level = dec_level;
	*hp = ip;
}

ridsym()
{
	struct ident **hp, *h;

	for( hp = symtab; hp < &symtab[SYMSIZE]; ++hp ){
		while( (h = *hp) != NULL && h->i_level > dec_level ){
			*hp = h->i_next;
			free(h);
		}
	}
}

/* define pre-declared things */
predecl()
{
	inputs = empty_group();
	outputs = empty_group();
	new_group("inputs", inputs);
	new_group("outputs", outputs);
}

/*
 * Action routines for parser
 */

/* Name *np is being declared as a signal ident, with indices lb..ub */
struct ident *
new_signal(np, lb, ub)
char *np;
int lb, ub;
{
	struct ident *ip;
	struct signal *sp;
	struct param *pp;
	struct expr **eq;
	int nb;

	ip = lookup(np);
	if( ip != NULL && ip->i_level >= dec_level ){
		yyerror("%s redeclared", np);
		free(np);
		return NULL;
	}
	ip = new(struct ident);
	ip->i_name = np;
	if( dec_level > 0 ){
		/* kludge - must be a parameter */
		pp = new(struct param);
		ip->i_type = PARAM;
		ip->i_obj.o_prm = pp;
		enter(ip);
		set_index(&pp->p_ix, lb, ub);
		pp->p_num = num_params;
		num_params += pp->p_ix.i_nel;
		return ip;
	}
	sp = new(struct signal);
	ip->i_type = SIGNAL;
	ip->i_obj.o_sig = sp;
	enter(ip);
	sp->s_ident = ip;
	init_attr(&sp->s_attr);
	set_index(&sp->s_ix, lb, ub);
	nb = sp->s_ix.i_nel * sizeof(struct expr *);
	eq = sp->s_def = (struct expr **)alloc(nb);
	for( nb = sp->s_ix.i_nel; nb > 0; --nb )
		*eq++ = NULL;
	sp->s_num = num_signals;
	num_signals += sp->s_ix.i_nel;
	return ip;
}

/* null attributes for newly defined signal */
init_attr(ap)
struct attr *ap;
{
	ap->a_bits = 0;
	ap->a_clock = NULL;
	ap->a_enable = NULL;
	ap->a_reset = NULL;
	ap->a_inclock = NULL;
}

#define SEEN_INPUT	0x10000
#define SEEN_OUTPUT	0x20000
#define SEEN_BITS	(SEEN_INPUT | SEEN_OUTPUT)

/* make an attribute block for attribute ab */
struct attr *
new_attr(ab, ex)
int ab;
struct expr *ex;
{
	struct attr *ap;

	ap = new(struct attr);
	ap->a_bits = ab;
	ap->a_clock = (ab == CLOCKED || ab == LATCHED)? ex: NULL;
	ap->a_enable = (ab == TRISTATE)? ex: NULL;
	ap->a_reset = (ab == ASYNC_RESET)? ex: NULL;
	ap->a_hazard = (ab == NO_HAZARD)? ex: NULL;
	ap->a_inclock = NULL;
	if (ab == INPUT_BIT)
	    ap->a_bits |= SEEN_INPUT;
	else if (ab == OUTPUT_BIT)
	    ap->a_bits |= SEEN_OUTPUT;
	return ap;
}

/* combine two attribute blocks */
struct attr *
add_attrs(a1, a2)
struct attr *a1, *a2;
{
	int i, j;

	if ((a2->a_bits & SEEN_BITS) == 0) {
	    if (a1->a_bits & SEEN_INPUT) {
		if (a2->a_bits == CLOCKED)
		    a2->a_bits = INPUT_CLOCK;
		else if (a2->a_bits == LATCHED)
		    a2->a_bits = INPUT_LATCH;
		if (a1->a_inclock == NULL)
		    a1->a_inclock = a2->a_clock;
		if (a2->a_bits & (TRISTATE | ASYNC_RESET | NO_HAZARD))
		    yyerror("only outputs can be enabled_by, reset_by or hazard_free");
	    } else {
		if (a1->a_clock == NULL)
		    a1->a_clock = a2->a_clock;
		if (a1->a_enable == NULL)
		    a1->a_enable = a2->a_enable;
		if (a1->a_reset == NULL)
		    a1->a_reset = a2->a_reset;
		if (a1->a_hazard == NULL)
		    a1->a_hazard = a2->a_hazard;
	    }
	} else {
	    if ((a1->a_bits & SEEN_BITS) == 0 && a2->a_bits & SEEN_INPUT) {
		if (a1->a_bits & CLOCKED)
		    a1->a_bits = (a1->a_bits & ~CLOCKED) | INPUT_CLOCK;
		if (a1->a_bits & LATCHED)
		    a1->a_bits = (a1->a_bits & ~LATCHED) | INPUT_LATCH;
		if (a1->a_clock != NULL) {
		    a1->a_inclock = a1->a_clock;
		    a1->a_clock = NULL;
		}
		if (a1->a_bits & (TRISTATE | ASYNC_RESET | NO_HAZARD)) {
		    yyerror("only outputs can be enabled_by, reset_by or hazard_free");
		    a1->a_bits &= ~(TRISTATE | ASYNC_RESET | NO_HAZARD);
		    a1->a_enable = NULL;
		    a1->a_reset = NULL;
		    a1->a_hazard = NULL;
		}
	    }
	    a1->a_bits = (a1->a_bits & ~SEEN_BITS) | (a2->a_bits & SEEN_BITS);
	}

	i = a1->a_bits | a2->a_bits;		/* combined attributes */
	j = a1->a_bits & a2->a_bits;		/* repeated attributes */
	if( (j & (CLOCKED | LATCHED | TRISTATE | ASYNC_RESET
		  | INPUT_CLOCK | INPUT_LATCH | NO_HAZARD)) != 0
	   || (~i & (HIGH_TRUE | LOW_TRUE)) == 0
	   || (~i & (INPUT_CLOCK | INPUT_LATCH)) == 0
	   || (~i & (CLOCKED | LATCHED)) == 0 ){
		yyerror("conflicting attributes");
		i = a1->a_bits;
	} else if( (j & ~(INPUT_BIT | OUTPUT_BIT | SEEN_BITS)) != 0 )
		yyerror("redundant attributes");
	a1->a_bits = i;
	free(a2);
	return a1;
}

/* index block for [0:0] */
struct index single = { 1, 0, 0, 1 };

/* set up an index structure for given bounds [lb:ub] */
set_index(ip, lb, ub)
struct index *ip;
int lb, ub;
{
	if( lb <= ub ){
		ip->i_nel = ub - lb + 1;
		ip->i_minix = lb;
		ip->i_maxix = ub;
		ip->i_dir = 1;
	} else {
		ip->i_nel = lb - ub + 1;
		ip->i_minix = ub;
		ip->i_maxix = lb;
		ip->i_dir = -1;
	}
}

/* Declare signal ident *ip to have attributes *ap */
signal_decl(ap, ip)
struct attr *ap;
struct ident *ip;
{
	struct signal *sp;

	if( ip == NULL )		/* was a redeclaration */
		return;
	sp = ip->i_obj.o_sig;
	sp->s_attr = *ap;
	if( (ap->a_bits & OUTPUT_BIT) != 0 )
		add_group(outputs, signal_expr(ip->i_name, -1, -1));
	else if( (ap->a_bits & INPUT_BIT) != 0 )
		add_group(inputs, signal_expr(ip->i_name, -1, -1));
}

/* Make a node for name *np, which must be a signal */
struct expr *
signal_expr(np, lb, ub)
char *np;
int lb, ub;
{
	struct ident *ip;
	struct expr *ep;
	int dir, nel, fx;
	struct signal *sp;
	struct function *fp;

	ip = lookup(np);
	if( ip == NULL ){
		yyerror("%s undefined", np);
		return NULL;
	}
	if( ip->i_type == GROUP ){
		ep = ip->i_obj.o_ex;
		if( lb >= 0 )
			ep = expr_select(ep, lb, ub);
	} else if( ip->i_type == SIGNAL ){
		sp = ip->i_obj.o_sig;
		indices(&sp->s_ix, lb, ub, &dir, &nel, &fx);
		ep = expand_group(ip->i_obj, SIGNAL, nel, dir, fx);
	} else {
		yyerror("%s is not a signal or group", np);
		ep = NULL;
	}
	return ep;
}

/* Make a leaf expression node for ident *np */
struct expr *
ident_expr(np, lb, ub)
char *np;
int lb, ub;
{
	struct ident *ip;
	struct expr *ep, *ee, *eq;
	int dir, nel, fx;
	struct signal *sp;
	struct function *fp;
	struct param *pp;
	struct state *stp;

	ip = lookup(np);
	if( ip == NULL ){
		yyerror("%s undefined", np);
		return bit_expr(Z);
	}
	switch( ip->i_type ){
	case SIGNAL:
		sp = ip->i_obj.o_sig;
		indices(&sp->s_ix, lb, ub, &dir, &nel, &fx);
		ep = expand_group(ip->i_obj, SIGNAL, nel, dir, fx);
		break;
	case PARAM:
		/* a formal parameter in a definition of a function */
		pp = ip->i_obj.o_prm;
		indices(&pp->p_ix, lb, ub, &dir, &nel, &fx);
		ep = expand_group(ip->i_obj, PARAM, nel, dir, fx);
		break;
	case GROUP:
		ep = ip->i_obj.o_ex;
		if( lb >= 0 )
			ep = expr_select(ep, lb, ub);
		break;
	case FUNCTION:
		/* function identifier used without any parameters */
		fp = ip->i_obj.o_func;
		if( fp->f_npar != 0 ){
			yyerror("function %s requires parameters", np);
			return NULL;
		}
		ep = fp->f_def;
		if( lb >= 0 )
			ep = expr_select(ep, lb, ub);
		break;
	case STATE:
		if( lb > 0 || ub > 0 )
			yyerror("index out of range");
		stp = ip->i_obj.o_state;
		ep = expr(EQUAL, stp->s_vars, stp->s_value);
		break;
	default:
		yyerror("%s can't be used in an expression: wrong type", np);
		ep = NULL;
	}
	return ep;
}

/* expand a multi-bit signal or parameter into a group */
struct expr *
expand_group(obj, type, nel, dir, fx)
union object obj;
int type, nel, dir, fx;
{
	struct expr *ep, *eq;
	int i;

	if( nel == 1 ){
		ep = new(struct expr);
		ep->e_op = type;
		ep->e_lu = obj;
		ep->e_index = fx;
	} else {
		ep = empty_group();
		for( i = 0; i < nel; ++i, fx += dir ){
			eq = new(struct expr);
			eq->e_op = type;
			eq->e_lu = obj;
			eq->e_index = fx;
			add_group(ep, eq);
		}
	}
	return ep;
}

/* Work out what range of indices are required */
indices(ixp, lb, ub, dir, nel, fx)
struct index *ixp;
int lb, ub, *dir, *nel, *fx;
{
	int t;

	if( lb == -1 ){		/* no indices specified - use 'em all */
		lb = ixp->i_minix;
		ub = ixp->i_maxix;
		*dir = ixp->i_dir;
	} else {
		if( lb > ub ){
			swap(lb, ub, t);
			*dir = -1;
		} else
			*dir = 1;
		if( lb < ixp->i_minix || ub > ixp->i_maxix ){
			yyerror("incompatible indices");
			if( lb < ixp->i_minix )
				lb = ixp->i_minix;
			if( ub > ixp->i_maxix )
				ub = ixp->i_maxix;
		}
	}
	*nel = ub - lb + 1;
	*fx = (*dir > 0)? lb: ub;
}

/* Construct an expression node with operator op and subtrees l & r */
struct expr *
expr(op, l, r)
int op;
struct expr *l, *r;
{
	struct expr *ep, *eg, *er;
	int nl, nr, o2, o3, v;

	if( op == ',' ){
		/* concatenate two groups */
		ep = make_group(l);
		add_group(ep, r);
		return ep;
	}
	nl = (l != NULL && l->e_op == GROUP)? l->e_nel: 1;
	nr = (r != NULL && r->e_op == GROUP)? r->e_nel: 1;
	if( nl != nr && !(op == '~' || (op == IF && nl == 1)) ){
		yyerror("number of elements does not match");
		return NULL;
	}
	if( nl == 1 && l != NULL && l->e_op == GROUP )
		l = l->e_first->e_left;
	if( nr == 1 && r != NULL && r->e_op == GROUP )
		r = r->e_first->e_left;
	if( nl > 1 ){
		l = l->e_first;
		if( r != NULL )
			r = r->e_first;
		switch( op ){
		case EQUAL:
		case NOTEQUAL:
			eg = NULL;
			o2 = (op == EQUAL)? '&': '|';
			for(; nl > 0; --nl ){
				ep = expr(op, l->e_left, r->e_left);
				eg = (eg == NULL)? ep: expr(o2, eg, ep);
				l = l->e_right;
				r = r->e_right;
			}
			break;
		case LESS:
		case LESSEQ:
		case GTR:
		case GTREQ:
			eg = NULL;
			er = NULL;
			o2 = (op == LESS || op == LESSEQ)? LESS: GTR;
			o3 = (o2 == LESS)? LESSEQ: GTREQ;
			for(; nl > 0; --nl ){
				ep = expr(o2, l->e_left, r->e_left);
				if( er != NULL )
					ep = expr('&', er, ep);
				eg = (eg == NULL)? ep: expr('|', eg, ep);
				ep = expr(o3, l->e_left, r->e_left);
				er = (er == NULL)? ep: expr('&', er, ep);
				l = l->e_right;
				r = r->e_right;
			}
			if( op == LESSEQ || op == GTREQ )
				eg = expr('|', eg, er);
			break;
		default:
			eg = empty_group();
			for(; nl > 0; --nl ){
				ep = new(struct expr);
				ep->e_op = op;
				ep->e_left = l->e_left;
				l = l->e_right;
				if( r != NULL ){
					ep->e_right = r->e_left;
					r = r->e_right;
				} else
					ep->e_right = NULL;
				add_group(eg, ep);
			}
		}
		return eg;
	}

	if( op == IF && nr > 1 ){
		/* if single then group else group */
		eg = empty_group();
		r = r->e_first;
		for(; nr > 0; --nr ){
			add_group(eg, expr(IF, l, r->e_left));
			r = r->e_right;
		}
		return eg;
	}

	if( l->e_op == BIT ){
		/* some optimizations */
		switch( l->e_bval ){
		case O:
			switch( op ){
			case '&':
			case EQUAL:
			case LESSEQ:
				return r;
			case '|':
				return l;
			case IF:
				return r->e_left;
			case NOTEQUAL:
			case GTR:
				return expr('~', r, NULL);
			case LESS:
			case '~':
				return bit_expr(Z);
			case GTREQ:
				return l;
			}
			break;
		case Z:
			switch( op ){
			case '&':
				return l;
			case '|':
			case NOTEQUAL:
			case LESS:
				return r;
			case IF:
				return r->e_right;
			case EQUAL:
			case GTREQ:
				return expr('~', r, NULL);
			case LESSEQ:
			case '~':
				return bit_expr(O);
			case GTR:
				return l;
			}
			break;
		case X:
			switch( op ){
			case '&':
			case '|':
			case '~':
				return l;
			case EQUAL:
			case LESSEQ:
			case GTREQ:
				return bit_expr(O);
			case NOTEQUAL:
			case LESS:
			case GTR:
				return bit_expr(Z);
			}
			break;
		}
	} else if( r != NULL && r->e_op == BIT ){
		switch( r->e_bval ){
		case O:
			switch( op ){
			case '&':
			case EQUAL:
			case GTREQ:
				return l;
			case '|':
				return r;
			case NOTEQUAL:
			case LESS:
				return expr('~', l, NULL);
			case LESSEQ:
				return bit_expr(O);
			case GTR:
				return bit_expr(Z);
			}
			break;
		case Z:
			switch( op ){
			case '&':
				return r;
			case '|':
			case NOTEQUAL:
			case GTR:
				return l;
			case EQUAL:
			case LESSEQ:
				return expr('~', l, NULL);
			case LESS:
				return bit_expr(Z);
			case GTREQ:
				return bit_expr(O);
			}
			break;
		case X:
			switch( op ){
			case '&':
			case '|':
				return r;
			case EQUAL:
			case LESSEQ:
			case GTREQ:
				return bit_expr(O);
			case NOTEQUAL:
			case LESS:
			case GTR:
				return bit_expr(Z);
			}
		}
	}

	ep = new(struct expr);
	ep->e_op = op;
	ep->e_left = l;
	ep->e_right = r;
	return ep;
}

/* Select elements from a group expression */
struct expr *
expr_select(ep, lb, ub)
struct expr *ep;
{
	int nel, dir;
	struct expr *eq;

	nel = ep->e_op == GROUP? ep->e_nel: 1;
	if( lb >= nel || ub >= nel ){
		yyerror("index out of range");
		if( lb >= nel )
			lb = nel - 1;
		if( ub >= nel )
			ub = nel - 1;
	}
	if( lb == ub ){
		if( ep->e_op == GROUP )
			eq = group_elt(ep, nel - lb - 1);
		else
			eq = ep;
	} else {
		eq = empty_group();
		dir = (ub > lb)? 1: -1;
		for(;;) {
			add_group(eq, group_elt(ep, nel - lb - 1));
			if( lb == ub )
				break;
			lb += dir;
		}
	}
	return eq;
}
	
/* Make a new, empty expression group */
struct expr *
empty_group()
{
	struct expr *ep;

	ep = new(struct expr);
	ep->e_op = GROUP;
	ep->e_nel = 0;
	ep->e_first = ep->e_last = NULL;
	return ep;
}

/* Add a new expression to a group */
add_group(grp, ep)
struct expr *grp, *ep;
{
	if( ep == NULL )
		return;
	if( ep->e_op == GROUP ){
		for( ep = ep->e_first; ep != NULL; ep = ep->e_right )
			add_group(grp, ep->e_left);
	} else {
		ep = make_elt(ep);
		if( grp->e_nel == 0 )
			grp->e_first = ep;
		else
			grp->e_last->e_right = ep;
		grp->e_last = ep;
		ep->e_right = NULL;
		++grp->e_nel;
	}
}

/* Make ep a group expression, if it isn't already */
struct expr *
make_group(ep)
struct expr *ep;
{
	struct expr *eq;

	if( ep != NULL && ep->e_op == GROUP )
		return ep;
	eq = empty_group();
	add_group(eq, ep);
	return eq;
}

/* Select one element of a group expression */
struct expr *
group_elt(gp, n)
struct expr *gp;
int n;
{
	struct expr *ep;

	if( gp == NULL || gp->e_op != GROUP )
		return (n == 0? gp: NULL);
	if( n < 0 || n >= gp->e_nel )
		return NULL;
	ep = gp->e_first;
	for(; n > 0 && ep != NULL; --n )
		ep = ep->e_right;
	return ep->e_left;
}

struct expr *
make_elt(ep)
struct expr *ep;
{
	struct expr *eq;

	while( ep->e_op == GRP_ELT )
		ep = ep->e_left;
	eq = new(struct expr);
	eq->e_op = GRP_ELT;
	eq->e_left = ep;
	eq->e_right = NULL;
	return eq;
}

struct expr *
copy_group(ep)
struct expr *ep;
{
	struct expr *eq;

	eq = empty_group();
	if( ep->e_op != GROUP )
		add_group(eq, ep);
	else
		for( ep = ep->e_first; ep != NULL; ep = ep->e_right )
			add_group(eq, ep->e_left);
	return eq;
}

/* Construct an expression node for a single bit b */
struct expr *
bit_expr(b)
int b;
{
	struct expr *ep;

	ep = new(struct expr);
	ep->e_op = BIT;
	ep->e_bval = b;
	return ep;
}

/* Construct an expression node for a number n of b bits */
struct expr *
hexnum_expr(n, b)
int n, b;
{
	struct expr *gp;

	gp = empty_group();
	while( --b >= 0 )
		add_group(gp, bit_expr((n & (1 << b)) != 0? O: Z));
	return gp;
}

/* Work out where definition of a variable is stored */
struct expr **
def_loc(vp)
struct expr *vp;
{
	struct signal *sp;

	if( vp == NULL || vp->e_op != SIGNAL ){
		yyerror("bug - not a signal");
		return NULL;
	}
	sp = vp->e_signal;
	return &sp->s_def[vp->e_index - sp->s_ix.i_minix];
}

/* Define variable *vp to have value *ep */
define_vble(vp, ep)
struct expr *vp, *ep;
{
	int i, nv, ne;

	if( vp == NULL || ep == NULL )
		return;
	nv = vp->e_op == GROUP? vp->e_nel: 1;
	ne = ep->e_op == GROUP? ep->e_nel: 1;
	if( nv != ne ){
	   	yyerror("incompatible number of bits in definition");
	   	return;
	}
	if( nv == 1 ){
		if( vp->e_op == GROUP )
			vp = vp->e_first->e_left;
		if( ep->e_op == GROUP )
			ep = ep->e_first->e_left;
		def_1_vble(vp, ep);
	} else {
		ep = ep->e_first;
		vp = vp->e_first;
		for( i = 0; i < nv; ++i ){
			if( vp == NULL || ep == NULL ){
				yyerror("bug - hit end of group (1)");
				return;
			}
			def_1_vble(vp->e_left, ep->e_left);
			vp = vp->e_right;
			ep = ep->e_right;
		}
	}
}

def_1_vble(vp, ep)
struct expr *vp, *ep;
{
	struct expr **dp;

	if( ddebug ){
		pexpr(vp);
		printf(" = ");
		pexpr(ep);
		printf("\n");
	}
	dp = def_loc(vp);
	if( dp != NULL ){
		if( *dp != NULL )
			expr_error("signal ", vp, " redefined");
		else
			*dp = ep;
	}
}

/* Define a new group called *np, with elements given in *ep */
new_group(np, ep)
char *np;
struct expr *ep;
{
	struct ident *ip;

	ip = lookup(np);
	if( ip != NULL ){
		yyerror("%s redeclared", np);
		free(np);
		return;
	}
	ip = new(struct ident);
	ip->i_name = np;
	ip->i_type = GROUP;
	ip->i_obj.o_ex = ep;
	enter(ip);
}

/* Declare a new level of symbol table entries (for function params) */
new_level()
{
	++dec_level;
	num_params = 0;
}

/* Declare another parameter for the current function */
/* tis all done in new_signal */
new_param(ip)
struct ident *ip;
{
}

/* Define function *np to be equal to expression *ep */
new_func(np, ep)
char *np;
struct expr *ep;
{
	struct ident *ip;
	struct function *fp;

	--dec_level;
	ridsym();
	ip = lookup(np);
	if( ip != NULL ){
		yyerror("%s redeclared", np);
		free(np);
		return;
	}
	ip = new(struct ident);
	fp = new(struct function);
	ip->i_name = np;
	ip->i_type = FUNCTION;
	ip->i_obj.o_func = fp;
	enter(ip);
	fp->f_npar = num_params;
	fp->f_def = ep;
}

/* Construct an expression for function *np applied to params *ep */
struct expr *
use_func(np, ep)
char *np;
struct expr *ep;
{
	struct ident *ip;
	struct function *fp;

	ip = lookup(np);
	if( ip == NULL ){
		yyerror("%s undefined", np);
		return NULL;
	}
	if( ip->i_type != FUNCTION ){
		yyerror("%s is not a function", np);
		return NULL;
	}
	ep = make_group(ep);
	fp = ip->i_obj.o_func;
	if( fp->f_npar < ep->e_nel )
		yyerror("too many parameters for function %s", np);
	else if( fp->f_npar > ep->e_nel )
		yyerror("too few parameters for function %s", np);
	return copy_func(fp->f_def, ep);
}

/* Copy the defining expression for function *fp, substituting params */
struct expr *
copy_func(ep, params)
struct expr *ep, *params;
{
	struct expr *eq, *p;
	struct param *pp;
	int pn;

	if( ep == NULL )
		return NULL;
	switch( ep->e_op ){
	case SIGNAL:
	case BIT:
		eq = new(struct expr);
		*eq = *ep;
		break;
	case PARAM:
		pp = ep->e_param;
		if( pp->p_ix.i_dir > 0 )
			pn = ep->e_index - pp->p_ix.i_minix;
		else
			pn = pp->p_ix.i_maxix - ep->e_index;
		eq = group_elt(params, pp->p_num + pn);
		break;
	case GROUP:
		eq = empty_group();
		for( p = ep->e_first; p != NULL; p = p->e_right )
			add_group(eq, copy_func(p->e_left, params));
		break;
	case ARRAY:
		eq = new(struct expr);
		eq->e_op = ARRAY;
		eq->e_array = ep->e_array;
                eq->e_left = copy_func(ep->e_left, params);
		break;
	default:
		eq = expr(ep->e_op, copy_func(ep->e_left, params),
				    copy_func(ep->e_right, params));
	}
	return eq;
}

/*
 * Procedures, etc. for logic arrays (truth tables and sum-of-products)
 */
struct expr *cur_logic;		/* group of logic arrays */
struct cube *cur_cube;		/* current cube being built up */
int cur_innum;			/* current input number */
int cur_outnum;			/* current output number */
struct expr *cur_logelt;	/* current element of cur_logic group */
int lgc_inputs;			/* total # of inputs in this array */
int lgc_outputs;		/* # elements in cur_logic */

/* Construct a truth table expression */
struct expr *
tt_expr()
{
	int i;
	struct expr *ep;

	ep = cur_logic->e_first;
	for( i = 0; i < cur_outnum; ++i ){
		check_tt(ep->e_left->e_array, i);
		ep = ep->e_right;
	}
	return cur_logic;
}

/* Construct a sum-of-products expression */
struct expr *
sp_expr()
{
	return cur_logic;
}

/* Make a new, empty bitlist */
new_bitlist()
{
	cur_innum = 0;
}

/* Add bit b to current bitlist */
add_bit(b)
int b;
{
	struct cube *cp;

	if( cur_innum >= 0 ){
		/* this is an input bit */
		if( cur_innum < lgc_inputs )
			/* what's more, it's a valid input bit */
			cur_cube->c_logic[cur_innum] = b;
		++cur_innum;
		return;
	}
	/* this is an output bit */
	if( cur_outnum < lgc_outputs ){
		cp = newcube(lgc_inputs + 1);
		ccopy(cp, cur_cube, lgc_inputs);
		cp->c_logic[lgc_inputs] = b;
		add(cur_logelt->e_left->e_array, cp);
		cur_logelt = cur_logelt->e_right;
	}
	++cur_outnum;
}

/* that's all the inputs, now for the outputs */
check_ins()
{
	if( cur_innum < lgc_inputs ){
		yyerror("too few input bits");
		for(; cur_innum < lgc_inputs; ++cur_innum )
			cur_cube->c_logic[cur_innum] = X;
	} else if( cur_innum > lgc_inputs )
		yyerror("too many input bits");
	cur_innum = -1;
	cur_outnum = 0;
	cur_logelt = cur_logic->e_first;
}

/* that's all the outputs */
check_outs()
{
	if( cur_outnum < lgc_outputs )
		yyerror("too few output bits");
	else if( cur_outnum > lgc_outputs )
		yyerror("too many output bits");
}

/* Make a new, empty truth table relation with *ins as the input bits and
   nouts outputs */
new_tt(ins, nouts)
struct expr *ins;
int nouts;
{
	struct expr *gp, *ig, *ep;
	int i;

	gp = empty_group();
	ig = make_group(ins);
	lgc_inputs = ig->e_nel;
	lgc_outputs = nouts;
	cur_cube = newcube(lgc_inputs + 1);
	for( i = 0; i < nouts; ++i ){
		ep = new(struct expr);
		ep->e_op = ARRAY;
		ep->e_left = ig;
		ep->e_array = newarray(lgc_inputs, lgc_inputs + 1);
		add_group(gp, ep);
	}
	cur_logic = gp;
}

add_tt()
{
}

/*
 * Procedures for state machines
 */
struct expr *mach_outs;			/* outputs of current state machine */
struct state *all_states;		/* all states referenced or defined */
struct expr *reset_trans;		/* state-independent transitions */
struct expr *dflt_trans;		/* transitions from unspecified states */

state_outputs(vp)
struct expr *vp;
{
	struct expr **dp;

	vp = make_group(vp);
	mach_outs = vp;
	/* check that none of the outputs is already defined */
	for( vp = vp->e_first; vp != NULL; vp = vp->e_right ){
		dp = def_loc(vp->e_left);
		if( dp != NULL && *dp != NULL )
			expr_error("signal ", vp, " redefined");
	}
	all_states = NULL;
	reset_trans = NULL;
	dflt_trans = NULL;
}

state_machine(async, states)
int async;
struct st_list *states;
{
	struct state *sp;
	int n;
	struct expr *ep, *vp, *p, **dp;

	if( async )
		yyerror("sorry, asynchronous state machines not implemented");
	for( sp = all_states; sp != NULL; sp = sp->s_next )
		if( sp->s_value == NULL ){
			yyerror("state %s undefined", sp->s_id->i_name);
			sp->s_value = bits_group(mach_outs->e_nel, X);
		}
	n = 0;
	for( p = mach_outs->e_first; p != NULL; p = p->e_right, ++n ){
		vp = p->e_left;
		ep = state_func(n);
		if( ddebug ){
			pexpr(vp);
			printf(" = ");
			pexpr(ep);
			printf("\n");
		}
		dp = def_loc(vp);
		if( dp != NULL && *dp == NULL )
			*dp = ep;
	}
}

struct state *
find_state(name)
char *name;
{
	struct ident *ip;
	struct state *sp;

	ip = lookup(name);
	if( ip == NULL ){
		/* new state, not yet seen */
		ip = new(struct ident);
		sp = new(struct state);
		ip->i_name = name;
		ip->i_type = STATE;
		ip->i_level = dec_level;
		ip->i_obj.o_state = sp;
		enter(ip);
		sp->s_vars = mach_outs;
		sp->s_value = NULL;		/* don't know yet */
		sp->s_trans = NULL;
		sp->s_next = all_states;
		all_states = sp;
		sp->s_id = ip;
	} else {
		/* existing name */
		if( ip->i_type != STATE ){
			yyerror("%s is not a state name", name);
			return NULL;
		}
		sp = ip->i_obj.o_state;
		if( sp->s_vars != mach_outs ){
			yyerror("%s is not a state in this state machine", name);
			return NULL;
		}
	}
	return sp;
}

state_indep(trans)
struct expr *trans;
{
	if( reset_trans != NULL )
		yyerror("all_states transitions already specified");
	else
		reset_trans = trans;
}

state_sig(name, sigs, trans)
char *name;
struct expr *sigs, *trans;
{
	struct expr *bits;

	bits = state_code(sigs, mach_outs);
	state_bit(name, bits, trans);
}

state_bit(name, bits, trans)
char *name;
struct expr *bits, *trans;
{
	struct state *sp;

	sp = find_state(name);
	if( sp == NULL )
		return;
	if( sp->s_value != NULL ){
		yyerror("state %s redefined", name);
		return;
	}
	bits = make_group(bits);
	if( bits->e_nel != mach_outs->e_nel ){
		yyerror("state %s: wrong number of output bits", name);
		bits = bits_group(mach_outs->e_nel, X);
	}
	sp->s_value = bits;
	sp->s_trans = trans;
}

state_dflt(trans)
struct expr *trans;
{
	if( dflt_trans != NULL )
		yyerror("other_states transitions already specified");
	else
		dflt_trans = trans;
}

struct expr *
goto_state(name)
char *name;
{
	struct state *sp;
	struct expr *ep;

	sp = find_state(name);
	ep = new(struct expr);
	ep->e_op = STATE;
	ep->e_state = sp;
	return ep;
}

struct expr *
state_code(grp, univ)
struct expr *grp, *univ;
{
	struct expr *bits, *bp, *ep, *p;
	int i;

	grp = make_group(grp);
	bits = bits_group(univ->e_nel, Z);	/* all outputs 0 to start */
	for( p = grp->e_first; p != NULL; p = p->e_right ){
		ep = p->e_left;
		if( ep->e_op != SIGNAL )
			yyerror("bug - not a signal (2)");
		else {
			i = st_outpos(ep, univ);
			if( i < 0 )
				expr_error("", ep,
				  " is not an output of this state machine");
			else {
				bp = group_elt(bits, i);
				if( bp->e_bval == O )
					expr_error("signal ", ep,
						"already included");
				else
					bp->e_bval = O;
			}
		}
	}
	return bits;
}

int
st_outpos(ep, univ)
struct expr *ep, *univ;
{
	int n;
	struct expr *up, *p;

	n = 0;
	for( p = univ->e_first; p != NULL; p = p->e_right, ++n ){
		up = p->e_left;
		if( up->e_op == SIGNAL && up->e_signal == ep->e_signal
		   && up->e_index == ep->e_index )
			return n;
	}
	return -1;
}

/* Make a group of n bits, each equal to val */
struct expr *bits_group(n, val)
int n, val;
{
	struct expr *grp;

	grp = empty_group();
	for( ; n > 0; --n )
		add_group(grp, bit_expr(val));
	return grp;
}

/* Return the function for state output n */
struct expr *
state_func(n)
int n;
{
	struct expr *ep, *ec, *et;
	struct state *sp;

	ep = bit_expr(X);
	if( dflt_trans != NULL )
		ep = trans_func(dflt_trans, n, ep);
	for( sp = all_states; sp != NULL; sp = sp->s_next ){
		ec = expr(EQUAL, sp->s_vars, sp->s_value);
		et = trans_func(sp->s_trans, n, group_elt(sp->s_value, n));
		ep = expr(IF, ec, expr(ELSE, et, ep));
	}
	if( reset_trans != NULL )
		ep = trans_func(reset_trans, n, ep);
	return ep;
}

/* Return a function for output n under a set of transitions */
struct expr *
trans_func(tr, n, dflt)
struct expr *tr, *dflt;
int n;
{
	struct expr *tp, *ep;

	if( tr == NULL )
		return dflt;
	switch( tr->e_op ){
	case STATE:
		return group_elt(tr->e_state->s_value, n);
	case IF:
		tp = trans_func(tr->e_right->e_left, n, dflt);
		ep = trans_func(tr->e_right->e_right, n, dflt);
		return expr(IF, tr->e_left, expr(ELSE, tp, ep));
	default:
		yyerror("bug - bad op in transitions (%d)\n", tr->e_op);
		return dflt;
	}
}

/*
 * Procedures to print out an expression in readable form
 */
expr_error(s1, ep, s2)
struct expr *ep;
char *s1, *s2;
{
	where();
	fprintf(stderr, "%s", s1);
	fpexpr(stderr, ep);
	fprintf(stderr, "%s\n", s2);
}

dump(ep)
struct expr *ep;
{
	pexpr(ep);
	printf("\n");
}

pexpr(ep)
struct expr *ep;
{
	fpexpr(stdout, ep);
}

fpexpr(f, ep)
FILE *f;
struct expr *ep;
{
	struct signal *sp;
	struct param *pp;
	int i;

	if( ep == NULL ){
		fprintf(f, "NULL");
		return;
	}
	switch( ep->e_op ){
	case SIGNAL:
		sp = ep->e_signal;
		fprintf(f, "%s", sp->s_ident->i_name);
		if( sp->s_ix.i_nel != 1 || sp->s_ix.i_minix != 0 )
			fprintf(f, "[%d]", ep->e_index);
		break;
	case GROUP:
		fprintf(f, "< ");
		i = ep->e_nel;
		ep = ep->e_left;
		for( ; i > 0; --i ){
			if( ep != NULL )
				fpexpr(f, ep->e_left);
			if( i > 1 )
				fprintf(f, ", ");
			if( ep != NULL )
				ep = ep->e_right;
		}
		if( ep != NULL )
			fprintf(f, ", ...");
		fprintf(f, " >");
		break;
	case BIT:
		i = ep->e_bval;
		if( i < N || i > X )
			fprintf(f, "#%d#", i);
		else
			fprintf(f, "%c", "?01%"[i]);
		break;
	case PIN:
		fprintf(f, "%d", ep->e_pin);
		break;
	case '~':
		fprop(f, ep->e_op);
		fpexpr(f, ep->e_left);
		break;
	case ARRAY:
		fprop(f, ep->e_op);
		fprintf(f, " ");
		fpexpr(f, ep->e_left);
                fprintf(f, " {\n");
		prary(ep->e_array, f);
		fprintf(f, "}");
		break;
	case PARAM:
		pp = ep->e_param;
		if( pp->p_ix.i_dir > 0 )
			i = ep->e_index - pp->p_ix.i_minix;
		else
			i = pp->p_ix.i_maxix - ep->e_index;
		fprintf(f, "param %d", pp->p_num + i);
		break;
	case IF:
		fprintf(f, "IF ");
		fpexpr(f, ep->e_left);
		fprintf(f, " THEN ");
		ep = ep->e_right;
		fpexpr(f, ep->e_left);
		fprintf(f, " ELSE ");
		fpexpr(f, ep->e_right);
		fprintf(f, " ENDIF");
		break;
	case STATE:
		fprintf(f, "state %s", ep->e_state->s_id->i_name);
		break;
	default:	/* assume binary operator */
		fprintf(f, "(");
		fpexpr(f, ep->e_left);
		fprintf(f, " ");
		fprop(f, ep->e_op);
		fprintf(f, " ");
		fpexpr(f, ep->e_right);
		fprintf(f, ")");
		break;
	}
}

fprop(f, op)
FILE *f;
int op;
{
	char *s;

	if( '!' <= op && op <= '~' )
		fprintf(f, "%c", op);
	else {
		switch( op ){
		case EQUAL:	s = "==";	break;
		case NOTEQUAL:	s = "!=";	break;
		case LESSEQ:	s = "<=";	break;
		case GTREQ:	s = ">=";	break;
		case LESS:	s = "<<";	break;
		case GTR:	s = ">>";	break;
		case IF:	s = "IF";	break;
		case ELSE:	s = "ELSE";	break;
		case ARRAY:	s = "ARRAY";	break;
		default:
			fprintf(f, "op%d", op);
			return;
		}
		fprintf(f, s);
	}
}

#if	FALSE
dump_defs(outs)
struct expr *outs;
{
	struct signal *sp;
	struct index *ixp;
	struct expr *ep, *p;
	int i;

	printf("Output definitions:\n");
	outs = make_group(outs);
	for( p = outs->e_first; p != NULL; p = p->e_right ){
		ep = p->e_left;
		sp = ep->e_signal;
		i = ep->e_index;
		ixp = &sp->s_ix;
		printf("%s", sp->s_ident->i_name);
		if( ixp->i_nel > 1 || ixp->i_minix != 0 )
			printf("[%d]", i);
		printf(" = ");
		pexpr(sp->s_def[i - ixp->i_minix]);
		printf("\n");
	}
}

/* Make a group expression out of the definitions of the outputs in outs. */
struct expr *
out_defs(outs)
struct expr *outs;
{
	struct signal *sp;
	struct index *ixp;
	struct expr *ep, *defs, *p;
	int i;

	outs = make_group(outs);
	defs = empty_group();
	for( p = outs->e_first; p != NULL; p = p->e_right ){
		ep = p->e_left;
		sp = ep->e_signal;
		i = ep->e_index;
		ixp = &sp->s_ix;
		add_group(defs, sp->s_def[i - ixp->i_minix]);
	}
	return defs;
}

#endif
